'''
Author: Octavian Suciu - osuciu@umiacs.umd.edu
Code Adjusted from http://www.github.com/embr/lsh
'''

import sys
import re
import json
from pyspark import SparkContext


FOLDER_IN = '/home/seed/Documents/spark/bin/data/documents'
FOLDER_OUT = '/home/seed/Documents/spark/bin/data/output_folder'
LSH_MODEL_FILE = '/home/seed/Documents/spark/bin/data/lsh_model'


'''
Pre-generated random numbers for persisting hash values used in building the model
'''
mm_pregenerated = [3626764237, 577090037, 4106135923, 1022050301, 1013818839, \
                2675342405, 3407369726, 1390851128, 973694259, 1988601454, \
                2454155475, 1942955373, 2038265566, 1112433019, 458825077, \
                4145683452, 1552728491, 2241903809, 778526663, 2908233303, \
                3889692776, 708452615, 4115477951, 3972265996, 3059489832, \
                1619040758, 3209591542, 2785274337, 485147340, 2354153409, \
                2315337690, 52734659, 332524003, 2449541944, 2271760180, \
                2356625382, 1411758002, 2929187291, 2746518602, 901304238, \
                1969700810, 1636471393, 2746317213, 165578901, 1754648232, \
                1167696004, 3815082342, 1511168161, 2354319937, 287264509, \
                2136903265, 1045900815, 4202002117, 2650602452, 3924066682, \
                388006705, 4148591305, 182818357, 2494333050, 962869285, \
                1322019967, 2123807367, 3985693436, 1911858404, 2045084184, \
                1781311907, 304122021, 321236752, 3185443825, 2938792530, \
                3907429125, 1390510041, 315365087, 1201798311, 3692505546, \
                1936365093, 1591528641, 3432174397, 3498157012, 624214061, \
                1166045467, 2172997514, 621570035, 2130974504, 3141902303, \
                857226971, 3399577998, 626871096, 1707201716, 348281753, \
                874757909, 358717904, 1808372858, 3918484893, 2348494799, \
                3264868694, 1591690432, 835531109, 1530718643, 1735072617]


# the number of bands
_b = 10
# the number of rows
_r = 10
# _b times _r
_n = 100
# the size of the n-grams
_shingle_size = 3


"""
This initializes the instance variable _memomask which is a list of the
random 32 bits associated with each hash function
"""
_memomask = mm_pregenerated

"""
This is a simple hash function which returns the result of a bitwise XOR
on the input x and the 32-bit random mask
"""
def _xor_hash(mask,x):
    return int(hash(x) ^ mask)


"""
Takes a sequence of tokenized words and maps each shingle to a unique id.
These unique ids, are then added to the shingle_vec object which is just a sparse
vector implemented as a dict with v[id]=1 when a shingle id is present
"""
def _get_shingle_vec(doc):
    sh = set()
    n = _shingle_size

    for j in range(0,len(doc) - n):
        s = doc[j:j+n]
        sh.add(s)

    return sh


"""
Takes a shingle vec and computes the minhash signature of length n using
approximate permutations.  This method is explained in Mining Massive
Datasets by Rajaraman and Ullman (http://infolab.stanford.edu/~ullman/mmds.html)
in section 3.3.4.
"""
def _get_sig(shingle_vec,num_perms):
    mhash = [sys.maxint for i in range(num_perms)]

    keys = sorted(shingle_vec)

    for r in keys:
        h = [_xor_hash(mask,r) % len(shingle_vec) for mask in _memomask]
        for i in range(num_perms):
            if (h[i] < mhash[i]):
                mhash[i] = h[i]

    return mhash


"""
Takes an n-dimensional minhash signature and computes b hashes for each of
b bands of r rows in the signature.  These hashes can take on any value that
can be stored in the 32bit integer.
"""
def _get_lsh(sig,b,r):
    lsh = []
    for i,band in enumerate(range(b)):
        #print tuple(sig[i*r:i*r+r])
        lsh.append(hash(tuple(sig[i*r:i*r+r])))
    return lsh


"""
Given an iterable of hashable items, returns a list of bucket ids
"""
def _get_lsh_bucket_ids_from_doc(doc):
    shingle_vec = _get_shingle_vec(doc)
    sig = _get_sig(shingle_vec,_n) # n-dimensional min-hash signiture
    lsh = _get_lsh(sig,_b,_r) # r-dimensional list of bucket ids

    return lsh


"""
Text normaization include removal of non-ASCII characters,
removal of extra white spaces and de-capitalizing letters
"""
def normalize_text(doc):
    tsplt = doc.split('\t')
    uid,txt = tsplt[:]

    txt = txt.strip().lower()
    txt = re.sub(r"[^\x00-\x7F]", '', txt)
    txt =  re.sub(r"\s+", ' ', txt)

    return [uid,txt]


"""
Interface for obtaining the LSH of a document
"""
def get_document_lsh(doc):
    uid,txt = doc

    lsh = _get_lsh_bucket_ids_from_doc(txt)

    return [uid,lsh]


"""
Obtainins the documents that match at least one entry from the model
"""
def get_document_matches(doc):
    uid,lsh = doc

    for i,lshval in enumerate(lsh):
        for e in lshModelBroadcast.value:
            if e[1][i] == lshval:
                return doc[0]

    return None


"""
Constructs the LSH model using a list of LSH entries
"""
def build_lsh_model(data):
    fout = open(LSH_MODEL_FILE,'w')
    for f in data[:-1]:
        fout.write('%s\n' % json.dumps(f))
    fout.write('%s' % json.dumps(data[-1]))


"""
Loads the pre-computed LSH model from a file
"""
def load_lsh_model():
    fie = open(LSH_MODEL_FILE)
    items = fie.read().split('\n')
    eidx = map(lambda x: json.loads(x),items)
    return eidx


"""
Main function.
Add your code here
"""
if __name__ == '__main__':
    #initializes the Spark Context
    sc = SparkContext()

    #loads the pre-built LSH model from disk
    lshModel = load_lsh_model()

    print "\n\n"
    # YOUR CODE BELOW
    # The print instructions help you debug each intermediate step
    # Do not modify the print output format.
    # Make sure you un-comment all the print instructions for the final submission


    # 1) load docsRDD from FOLDER_IN
    docsRDD = None
    #print "1> Load data from file\t[...]\n"

    # 2) create a broadcast variable lshModelBroadcast containing the LSH model
    lshModelBroadcast = None
    #print "2> Create broadcast variable\t[...]\n"

    # 3) normalize the text in the documents
    normalizedDocsRDD = None
    #print "3> Normalize text in input documents\t[...]\n"

    # 4) compute the LSH for the normalized documents
    lshRDD = None
    #print "4> Compute LSH\t%s\n" % lshRDD.collect()

    # 5) compute the matches with the model for all the documents
    matchesRDD = None
    #print "5> Compute matches\t%s\n" % matchesRDD.collect()

    # 6) remove all the 'None' entries from the matches - which correspond to documents without a match in the model
    candidatesRDD = None
    #print "6> Remove al None entries\t%s\n" % candidatesRDD.collect()

    # 7) save candidatesRDD to FOLDER_OUT.
    # NOTE: delete FOLDER_OUT from disk between consecutive runs of the program
    pass
    #print "7> Save results to the output folder\t[...]\n",

    # END OF EDITS